import numpy as np
import csv

class PCTCurve:
    """Simple representation of a PCT curve."""

    @classmethod
    def from_csv(cls, temperature: float, csv_path: str = None, csv_file = None):
        """
        Given the path to a CSV file, instantiate a PCTCurve assuming that the
        first column in the CSV is a list of concentrations, and the second
        column is a list of pressures.

        :param csv_path: A file path to the CSV containing the PCTCurve
        :param temperature: The temperature at which the measurement was taken
        """
        concentrations = []
        pressures = []

        if csv_path is not None:
            with open(csv_path, "r", newline='') as csv_file:
                reader = csv.reader(csv_file)
                for row in reader:
                    concentrations.append(float(row[0]))
                    pressures.append(float(row[1]))
        elif csv_file is not None:
            reader = csv.reader(csv_file)
            for row in reader:
                concentrations.append(float(row[0]))
                pressures.append(float(row[1]))

        return cls(concentrations, pressures, temperature)

    def __init__(self, concentrations: list, pressures: list, temperature: float):
        self.concentrations = np.array(concentrations)
        self.pressures = np.array(pressures)
        self.temperature = temperature

    def get_closest_vant_hoff_point(self, target_concentration: float):
        """
        Given a target concentration, finds the nearest measured point on
        the curve, and then returns a van't hoff point associated with
        that measured point
        """
        distances = np.abs(self.concentrations - target_concentration)
        closest_point_idx = np.argmin(distances)
        pressure = self.pressures[closest_point_idx]
        return np.array([1 / self.temperature, np.log(pressure)])

    def interpolate_point_at_concentration(self, concentration: float):
        distances = self.concentrations - concentration
        above = np.where(distances > 0)[0][0]
        below = np.where(distances <= 0)[0][-1]
        total_difference = self.concentrations[above] - self.concentrations[below]
        weighted_upper_pressure = self.pressures[above] * (concentration - self.concentrations[below])
        weighted_lower_pressure = self.pressures[below] * (self.concentrations[above] - concentration)
        avg_pressure = ( weighted_upper_pressure + weighted_lower_pressure ) / total_difference
        return np.array([1 / self.temperature, np.log(avg_pressure)])

    def get_vant_hoff_values_at_idx(self, idx: int):
        """
        Given the index of a point, returns the van't hoff
        values at that point.
        """
        pressure = self._pressures[idx]
        return np.array([1 / self.temperature, np.log(pressure)])



import vant_hoff_analysis as vh
import numpy as np


CSV_PATH = "tests/data/des1_348K.csv"

def test_pct_get_point():
    concentrations = [1, 2.4, 3.6, 4.8, 5.9]
    pressures = [1.1, 2.2, 3.3, 5.5, 6.7]
    temperature = 100
    curve = vh.PCTCurve(concentrations, pressures, temperature)
    pt = curve.get_closest_vant_hoff_point(4)
    assert pt[0] == 1 / 100
    assert pt[1] == np.log(3.3)


def test_curve_a():
    curve_a_pressures = [np.exp(-1), np.exp(0), np.exp(1)]
    curve_a_concs = [0.2, 0.52, 0.85]
    curve_a_temperature = 278

    curve_a = vh.PCTCurve(curve_a_concs, curve_a_pressures, curve_a_temperature)
    pt = curve_a.get_closest_vant_hoff_point(0.5)
    assert pt[0] == 1/278
    assert pt[1] == 0

def test_curve_b():
    curve_b_pressures = [np.exp(3), np.exp(4), np.exp(5)]
    curve_b_concs = [0.05, 0.49, 0.95]
    curve_b_temperature = 373

    curve_b = vh.PCTCurve(curve_b_concs, curve_b_pressures, curve_b_temperature)
    pt = curve_b.get_closest_vant_hoff_point(0.5)
    assert pt[0] == 1/373
    assert pt[1] == 4

def test_from_csv():
    curve = vh.PCTCurve.from_csv(348, csv_path=CSV_PATH)
    assert curve.temperature == 348
    assert curve.concentrations[0] == 0.8309570766512838
    assert curve.pressures[0] == 0.04403285286053071

def test_from_csv_pointer():
    with open(CSV_PATH) as f:
        curve = vh.PCTCurve.from_csv(348, csv_file=f)
        assert curve.temperature == 348
        assert curve.concentrations[0] == 0.8309570766512838
        assert curve.pressures[0] == 0.04403285286053071



import numpy as np
from scipy import constants
from typing import List
from .pct_curve import PCTCurve
from .result import VHResult

# Steps:
# 1. From each PCT curve, note temperature, find the point nearest
# the provided X value, and grab it's pressure
# 2. For each point, convert temperature to 1/T (units 1/K) and pressure to ln(P),
# 3. Perform a regression to find line of best fit for those points.
# 4. the deltaH value will be m * R
# 5. the deltaS value will be -b * R

class Fitter:

    def fit_at_concentration(self, concentration, curves: List[PCTCurve]):
        """Given a concentration, identifies the nearest points in each curve to
        that concentration and then performs a Van't Hoff fit on those points.
        """
        points = np.array(
            list(map(lambda curve: curve.interpolate_point_at_concentration(concentration), curves)))
        xs = points.T[0]
        ys = points.T[1]
        return self.fit(xs, ys)

    def fit_from_indices(self, indices: List[int], curves: List[PCTCurve]):
        """
        Given a list of curves, and a list of indices which indicate the points on
        the curves which should be used for analysis, return the thermodynamic values
        resulting from performing that analysis.

        :param indices: A list of indices
        :param curves: The PCT c
        """
        assert len(indices) == len(curves), "Must provide one index for each curve"
        points = [(c.temperature, c.pressures[idx]) for (c, idx) in zip(curves, indices)]
        return self.fit_from_points(points)

    def fit_from_points(self, points):
        """Given (temperature, pressure) points, returns thermodynamic fit data"""
        xs = np.array([1 / point[0] for point in points])
        ys = np.array([np.log(point[1]) for point in points])
        # Check for values that resulted in NaN points - e.g. negative pressure values, which
        # invalid as inputs to logarithm.
        # probably a more elegent way to do this...
        if np.any(np.isnan(xs)) or np.any(np.isnan(ys)):
            raise ValueError("Invalid inputs detected")
        return self.fit(xs, ys)

    def fit(self, xs, ys):
        """Given 1/T, ln(pressure) points, perform a linear fit, and then
        return the thermodynamic constants.

        :param xs: A list of 1/temperature values
        :param ys: A list of log pressure values
        """
        assert len(xs) == len(ys), "x and y lists must have the same length"
        b, m = np.polynomial.polynomial.polyfit(xs, ys, deg=1)
        deltaH = -m * constants.R / 1000
        negative_deltaS = -b * constants.R
        return VHResult(deltaH, -negative_deltaS, xs.tolist(), ys.tolist(), m, b)





class VHResult():

    def __init__(self, deltaH, deltaS, xs, ys, m, b):
        self.deltaH = deltaH
        self.deltaS = deltaS
        self.xs = xs
        self.ys = ys
        self.m = m
        self.b = b

    def to_dict(self):
        return {
            "deltaH": self.deltaH,
            "deltaS": self.deltaS,
            "m": self.m,
            "b": self.b,
            "xs": self.xs,
            "ys": self.ys,
        }


import scipy
import numpy as np
import pytest
from vant_hoff_analysis import Fitter, PCTCurve

@pytest.fixture
def curve_a():
    curve_a_pressures = [np.exp(-1), np.exp(0), np.exp(1)]
    curve_a_concs = [0.2, 0.52, 0.85]
    curve_a_temperature = 278

    curve_a = PCTCurve(curve_a_concs, curve_a_pressures, curve_a_temperature)
    return curve_a

@pytest.fixture
def curve_b():
    curve_b_pressures = [np.exp(3), np.exp(4), np.exp(5)]
    curve_b_concs = [0.05, 0.49, 0.95]
    curve_b_temperature = 373

    curve_b = PCTCurve(curve_b_concs, curve_b_pressures, curve_b_temperature)
    return curve_b

def test_fitter(curve_a, curve_b):
    # A contrived example of fit
    fitter = Fitter()
    fit_result = fitter.fit_at_concentration(0.5, [curve_a, curve_b])
    delta_H = fit_result.to_dict()["deltaH"]
    delta_S = fit_result.to_dict()["deltaS"]

    assert delta_H == pytest.approx(4.450090593 * scipy.constants.R)
    assert delta_S == pytest.approx(15.96721097 * scipy.constants.R)

def test_from_points():
    point1 = [348, 0.193889213]
    point2 = [373, 0.563732355]
    point3 = [388, 1.061087193]

    fitter = Fitter()

    fit_result = fitter.fit_from_points([point1, point2, point3])
    delta_H = fit_result.to_dict()["deltaH"]
    delta_S = fit_result.to_dict()["deltaS"]

    assert delta_H == pytest.approx(47.4999715)
    assert delta_S == pytest.approx(122.783428)

def test_nan_case():
    points = [[2, -10], [3, 3]]

    fitter = Fitter()

    with pytest.raises(ValueError):
        fitter.fit_from_points(points)

def test_from_indices(curve_a, curve_b):
    fitter = Fitter()
    fit_result = fitter.fit_from_indices([1,1], [curve_a, curve_b])
    delta_H = fit_result.to_dict()["deltaH"]
    delta_S = fit_result.to_dict()["deltaS"]

    assert delta_H == pytest.approx(4.36606315789 * scipy.constants.R)
    assert delta_S == pytest.approx(15.7052631579 * scipy.constants.R)


from setuptools import setup

setup(
    name="vant_hoff_analysis",
    version="0.0.2",
    description="A small tool for performing van't Hoff analysis on PCT hydrogenation data",
    author="Max Gallant",
    packages=["vant_hoff_analysis", "tests"],
    install_requires=["numpy==1.20.2", "scipy==1.6.3"]
)